#ifndef LIST_H
#define LIST_H

#ifndef LIST_MAX
#define LIST_MAX (128)
#endif

enum list_insert_type { NO_INSERT_FIRST = 1, NO_INSERT_MIDDLE = 2, NO_INSERT_LAST = 4 };
enum list_delete_type { NO_DELETE_FIRST = 8, NO_DELETE_MIDDLE = 16, NO_DELETE_LAST = 32 };
enum linked_type { DOUBLY_LINKED, SINGLY_LINKED = 64, NON_CIRCULAR, CIRCULAR = 128 };

#define STACK_TYPE (54)
#define QUEUE_TYPE (51)
#define LINKED_TYPE (0)

#define array(type) ((((type) / NO_INSERT_MIDDLE) % 2) && (((type) / NO_DELETE_MIDDLE) % 2))
#define single(type) (((type) / SINGLY_LINKED) % 2)

struct single_node {
	void *element;
	struct single_node *next;
};

struct double_node {
	void *element;
	struct double_node *previous, *next;
};

union linked_node {
	struct single_node *single;
	struct double_node *double_;
};

struct linked_base {
	union linked_node base, avail;
};

union list_base {
	void *array;
	struct linked_base linked;
};

union list_end {
	struct array_end {
		void *first, *last;
	} array;
	struct linked_end {
		union linked_node first, last;
	} linked;
};

struct list {
	int type, size, len;
	union list_base base;
	union list_end end;
	visit_t visit;
	bool_t inuse;
};

union list_node {
	struct list list;
	union list_node *next;
};

struct list_pool {
	union list_node base[LIST_MAX], *avail;
	lock_t free;
};

typedef const struct list * const list_t;

struct linked_ptr {
	union linked_node current, previous;
};

typedef union list_ptr {
	void *array;
	struct linked_ptr linked;
} listptr_t;

listptr_t *listptr_init(listptr_t *ptr, int type);

list_t list_new(int type, int size);
void list_dispose(list_t list);

list_t list_insert(list_t list, void *element, listptr_t *ptr);
list_t list_append(list_t list, void *element, listptr_t *ptr);
list_t list_delete(list_t list, listptr_t *ptr);
list_t list_execute(list_t list, listptr_t *ptr);

void *list_get(list_t list, int get, listptr_t *ptr);

list_t *list_resize(list_t *list, int size);

#define list_type(list) ((list)->type)
#define list_size(list) ((list)->size)
#define list_count(list) ((list)->count)

#endif
