
#include"partial.h"

#include"misc.h"

unsigned partial_size(unsigned size, unsigned short count) {
	return size * isqrt(count) * 2;
}

struct partial_table *partial_new(struct partial_table *table, unsigned table_size, unsigned size, unsigned short count) {
	static struct partial_table *new;

	if((new = table) ||(new = malloc(PARTIAL_TABLE_SIZE)) {
		static struct partial *base;

		if(table_size <= 0) table_size = partial_size(size, count);
		if(base = malloc(new->size = PARTIAL_SIZE * table_size)) {
			int partial;

			for(partial = 0; partial < table_size;) {
				if(partial < size) {
					base[partial].partial = partial;
					base[partial].count = count;
				}
				base[partial].next = base + ++partial;
			}
			base[size - 1].next = NULL;
			new->free = base;
			new->avail = base + size;
			new->mutex.avail = new->mutex.free = PTHREAD_MUTEX_INITIALIZER;
			return new;
		}
		if(! table) free(new);
	}
	return NULL;
}

void partial_dispose(struct partial_table *table, int free) {
	free(table->base);
	if(free) free(table);
	return;
}

struct partial *partial_extend(struct partial_table *table, struct partial *head, unsigned short count) {
	struct partial *last, *partial, *free, *extend;
	unsigned short remaining;

	if(extend = last = head) {
		while(last->next) last = last->next;
	}
	free = previous = NULL;
	pthread_mutex_lock(&table->mutex.free);
	least = table->least;
	while(least) {
		if(least->partial == last->partial && least->offset == last->offset + last->count) {
			free = least;
			break;
		}
		previous = least;
		least = least->next;
	}
	remaining = count;
	while(remaining) {
		if(! free) {
			free = table->least;
			table->least = table->least->next
		}
		else {
			if(previous) previous->next = free->next;
			else table->least = free->next;
		}
		free->next = NULL;
		if(most) most->next = free;
		most = free;
		if(remaining >= free->count) {
			table->free = free->next;
			pthread_mutex_unlock(&table->mutex.free);
			partial = free;
			remaining -= free->count;
		}
		else {
			pthread_mutex_lock(&table->mutex.avail);
			if(! partial = table->avail) {
				pthread_mutex_unlock(&table->mutex.avail);
				partial_trim(table, head, count - remaining);
				return NULL;
			}
			table->avail = table->avail->next;
			pthread_mutex_unlock(&table->mutex.avail);
			*partial = *free;
			free->offset += remaining;
			free->count -= remaining;
			pthread_mutex_unlock(&table->mutex.free);
			partial->count = remaining;
			remaining = 0;
		}
		if(! extend) {
			extend = partial;
		}
		else {
			if(last->partial == partial->partial && last->offset + last->count == partial->offset) {
				last->count += partial->count;
				pthread_mutex_lock(&table->mutex.avail);
				partial->next = table->avail;
				table->avail = partial;
				pthread_mutex_unlock(&table->mutex.avail);
				partial = last;
			}
		}
		last = last->next = partial;
	}
	last->next = NULL;
	return extend;
}

#ifdef PARTIAL_LEN
#define _PARTIAL_LEN PARTIAL_LEN
#undef PARTIAL_LEN
#define PARTIAL_LEN (512)
#endif

struct partial *list[PARTIAL_LEN];

int partial_trim(struct partial_table *table, struct partial *head, unsigned short count) {
	struct partial **entry, *partial;
	unsigned short trim;

	if(partial = head) {
		trim = count;
		entry = list;
		do {
			if(entry - list >= PARTIAL_LEN) return 0;
			*entry = partial;
			if(! count) trim += partial->count;
			partial = partial->next;
			entry++;
		} while(partial);
		while(trim && entry > list) {
			struct partial *free, *previous;

			--entry;
			if(trim >= *entry->count) {
				trim -= *entry->count;
			}
			else {
				struct partial *new;

				pthread_mutex_lock(&table->mutex.avail);
				if(new = table->avail) {
					*entry->count -= trim;
					table->avail = table->avail->next;
					pthread_mutex_unlock(&table->mutex.avail);
					*new = **entry;
					new->offset += new->count;
					new->count = trim;
					*entry = new;
					trim = 0;
				}
				else {
					pthread_mutex_unlock(&table->mutex.avail);
					return 0;
				}
			}
			pthread_mutex_lock(&table->mutex.free);
			if(free = table->least) {
				while(free) {
					if(free->partial == *entry->partial) {
						if(free->offset == *entry->offset + *entry->count) {
							free->offset = *entry->offset;
							free->count += *entry->count;
							pthread_mutex_lock(&table->mutex.avail);
							*entry->next = table->avail;
							table->avail = *entry;
							pthread_mutex_unlock(&table->mutex.avail);
						}
						else {
							if(free->offset + free->count == *entry->offset) {
								free->count += *entry->count;
								pthread_mutex_lock(&table->mutex.avail);
								*entry->next = table->avail;
								table->avail = *entry;
								pthread_mutex_unlock(&table->mutex.avail);
							}
						}
					}
					free = free->next;
				}
			}
			if(previous) {
				free = previous;
				*entry->next = free->next;
				free->next = *entry;
			}
			else {
				table->free = *entry;
				*entry->next = free;
				free = *entry;
				if(free->next) *entry = free->next;
			}
			while(free->partial == *entry->partial && free->offset + free->count == *entry->offset) {
				free->count += *entry->count;
				free->next = *entry->next;
				pthread_mutex_lock(&table->mutex.avail);
				*entry->next = table->avail;
				table->avail = *entry;
				pthread_mutex_unlock(&table->mutex.avail);
				*entry = free->next;
			}
			pthread_mutex_unlock(&table->mutex.free);
		}
	}
	return 1;
}

#ifdef _PARTIAL_LEN
#undef PARTIAL_LEN
#define PARTIAL_LEN _PARTIAL_LEN
#undef _PARTIAL_LEN
#endif
