/*

    List management

    Copyright (C) 2006  Sundaresh Venugopal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    email : sundaresh@mail.com

    Address : K-222, Usha's appartments, 16-th main, Jayanagar 4-th block, Bangalore - 560011

    Phone : 91-80-26651501

*/
#ifndef LIST_H
#define LIST_H

#ifndef LIST_MAX
#define LIST_MAX (128)
#endif

enum list_insert_type { NO_INSERT_FIRST = 1, NO_INSERT_MIDDLE = 2, NO_INSERT_LAST = 4 };
enum list_delete_type { NO_DELETE_FIRST = 8, NO_DELETE_MIDDLE = 16, NO_DELETE_LAST = 32 };
enum linked_type { DOUBLY_LINKED, SINGLY_LINKED = 64, NON_CIRCULAR, CIRCULAR = 128 };

#define STACK_TYPE (54)
#define QUEUE_TYPE (51)
#define LINKED_TYPE (0)

struct single_node {
	void *element;
	struct single_node *next;
};

struct double_node {
	void *element;
	struct double_node *previous, *next;
};

union linked_node {
	struct single_node *single;
	struct double_node *double_;
};

struct linked_base {
	union linked_node base, avail;
};

union list_base {
	void *array;
	struct linked_base linked;
};

union list_end {
	struct array_end {
		void *first, *last;
	} array;
	struct linked_end {
		union linked_node first, last;
	} linked;
};

struct list {
	int type, size, len;
	union list_base base;
	union list_end end;
	counter_t counter;
	bool_t inuse;
};

typedef const struct list * const list_t;

struct linked_ptr {
	union linked_node current, previous;
};

typedef union list_ptr {
	void *array;
	struct linked_ptr linked;
} listptr_t;

listptr_t *listptr_init(listptr_t *ptr, int type);

list_t list_new(int type, int size);
void list_dispose(list_t list);

list_t list_insert(list_t list, void *element, listptr_t *ptr);
list_t list_append(list_t list, void *element, listptr_t *ptr);
list_t list_delete(list_t list, listptr_t *ptr);
list_t list_execute(list_t list, listptr_t *ptr);

void *list_get(list_t list, int get, listptr_t *ptr);

list_t list_resize(list_t list, int size);

#define list_type(list) ((list)->type)
#define list_size(list) ((list)->size)
#define list_count(list) ((list)->count)

#endif
