/*

    Table Management Module

    Copyright (C) 2006  Sundaresh Venugopal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    email : sundaresh@mail.com

    Address : K-222, Usha's appartments, 16-th main, Jayanagar 4-th block, Bangalore - 560011

    Phone : 91-80-26651501

*/
#ifndef TABLE_H
#define TABLE_H

#include"lock.h"
#include"counter.h"

#include"get.h"
#include"compare.h"

/* Maximum number of tables per process */
#define TABLE_MAX (64)

/* Maximum number of distinct table types */
#define TABLE_TYPE_MAX (16)

enum table_duplicates { NO_DUPLICATES, DUPLICATES = 8 };
typedef enum table_type { LIST_TABLE = 16, TREE_TABLE = 0, HASH_TABLE = 128 } tabletype_t;

typedef struct table_base {
	void *data, *avail;
	union {
		struct binary_base binary;
		struct hash_base hash;
		struct list_base list;
	} table;
} tablebase_t;

typedef union table_parameter {
	struct balance {
		int sum, ratio;
	} balance;
	struct hash {
		float occupancy;
		union  {
			hash_fn hash, rehash;
		} fn;
	} hash;
} tableparameter_t;

typedef struct table {
	int type;
	int size[3];
	int count[3];
	tablebase_t base;
	compare_fn compare;
	tableparameter_t parameter;
	lock_t free;
	counter_t counter;
	bool_t inuse;
} const * const table_t;

typedef const struct table * const table_t;

typedef union table_ptr {
} tableptr_t;

struct table_op {
	struct table *(*new)(struct table *table);
	void (*dispose)(struct table *table);
	void *(*lookup)(struct table *table, void *element, union table_ptr *ptr, int lookup);
	void *(*get)(struct table *table, union table_ptr *ptr, int get);
	struct table *(*resize)(struct table *table, int size);
};

table_t table_new(int type, int size, int width, compare_fn compare, hash_fn hash_rehash, float occupancy);
void table_dispose(table_t table);

void *table_lookup(table_t table, void *element, tableptr_t *ptr, int lookup);

void *table_get(table_t table, void *element, tableptr_t *ptr, int get);

table_t table_resize(table_t table, int size);

#endif
