#include<stddef.h>
#include<stdlib.h>

#include"lock.h"
#include"mutex.h"
#include"cache.h"

static struct cache_entry *adjust_entry(struct cache *cache, struct cache_entry *entry) {
	if(entry->previous) {
		while(1) {
			if(lock(&entry->previous->lock) > 0) {
				if(lock(&entry->lock) > 0) {
					entry->previous = NULL;
					entry->previous->next = entry->next;
					unlock(&entry->previous->lock);
					break;
				}
				unlock(&entry->previous->lock);
			}
			pthread_skip();
		}
		if(entry->next) {
			unlock(&entry->lock);
			while(1) {
				if(lock(&entry->next->lock) > 0) {
					if(lock(&entry->lock) > 0) {
						entry->next->previous = entry->previous;
						unlock(&entry->next->lock);
						break;
					}
					unlock(&entry->next->lock);
				}
				pthread_skip();
			}
		}
		else {
			cache->table.least = cache->table.least->previous;
			cache->table.least->next = NULL;
		}
		pthread_mutex_lock(&cache->table.mutex.head);
		entry->next = cache->table.head;
		cache->table.head->previous = cache->table.head = entry;
		pthread_mutex_unlock(&cache->table.mutex.head);
	}
	unlock(&entry->lock);
	return entry;
}

void *cache_read(struct cache *cache, long address, void *buffer) {
	struct cache_entry *entry;

	pthread_mutex_lock(&cache->table.mutex.root);
	entry = cache->table.root;
	pthread_mutex_unlock(&cache->table.mutex.root);
	while(1) {
		struct cache_entry *child;

		if(lock(&entry->lock) > 0) {
			if(entry->block.address != address) {
				child = (entry->block.address < address) ? entry->right : entry->left;
				if(child) {
					unlock(&entry->lock);
					entry = child;
					continue;
				}
			}
			break;
		}
		pthread_skip();
	}
	if(entry->block.address == address) {
		adjust_entry(cache, entry);
		/* pcp is physical copy, buffer is a physical address, a RAM location;
		 * while ptr is a virtual address. The opposite of pcp is vcp. */
		return pcp(buffer, entry->node.ptr, cache->fs->size.block);
	}
	unlock(&entry->lock);
	return NULL;
}

static void swap_node(struct cache_entry *entry, struct cache_entry *previous, struct block *block) {
	struct block temp;

	temp = entry->block;
	if(temp.previous) {
	}
	entry->block = *block;
	*block = temp;
	if(previous) {
		while(1) {
			if(lock(&previous->lock) > 0) {
				if(lock(&previous->previous->lock) > 0) {
					if(lock(&previous->next->lock) > 0) {
						previous->previous->next = entry;
						previous->next->previous = entry;
						unlock(&previous->lock);
						unlock(&previous->previous->lock);
						unlock(&previous->next->lock);
						break;
					}
					unlock(&previous->previous->lock);
				}
				unlock(&previous->lock);
			}
			pthread_skip();
		}
	}
	return;
}

struct cache *cache_write(struct cache *cache, long address, void *buffer) {
	struct cache_entry *entry, *parent, *lparent, *sparent, *successor, *pivotal, *new, *previous;
	struct cache_entry_node node;
	unsigned key, bit, first;

	first = 0;
	node.address = address;
	node.ptr = NULL;
	pivotal = previous = NULL;
	key = bit = 1 << (INT_BITS - 1);
	parent = entry = cache->table.root;
	while(entry) {
		bit >>= 1;
		if(first != 1) unlock(&parent->lock);
		if(lock(&entry->lock) <= 0) pthread_skip();
		parent = entry;
		if(entry->node.address == node.address) break;
		if(entry->node.address < node.address && (node.address & key)) {
			entry = entry->right;
			key |= bit;
		}
		else {
			if(entry->node.address > node.address && (! (node.address & key))) {
				entry = entry->left;
				key >>= 1;
			}
			else {
				swap_node(entry, previous, &node);
				previous = entry;
				if(++first == 1) pivotal = parent;
			}
		}
	}
	if(! entry) {
		while(lock(&cache->table.least) <= 0) pthread_skip();
		new = cache->table.least;
		pthread_mutex_lock(&cache->table.mutex.root);
		entry = cache->table.root;
		pthread_mutex_unlock(&cache->table.mutex.root);
		lparent = NULL;
		while(entry != new) {
			if(lock(&entry->lock) > 0) {
				if(lparent) unlock(&lparent->lock);
				lparent = entry;
				entry = (entry->node.address < least->node.address) ? entry->right : entry->left;
			}
			pthread_skip();
		}
		cache->table.least = new->previous;
		sparent = NULL;
		successor = new->right;
		while(1) {
			if(lock(&successor->lock) > 0) {
				if(successor->left) {
					if(sparent) unlock(&sparent->lock);
					sparent = successor;
					successor = successor->left;
				}
				else break;
			}
			pthread_skip();
		}
		if(lparent) {
			if(lparent->left == new) lparent->left = successor;
			else lparent->right = successor;
		}
		else {
			pthread_mutex_lock(&cache->table.mutex.root);
			cache->table.root = successor;
			pthread_mutex_unlock(&cache->table.mutex.root);
		}
		if(sparent) sparent->left = successor->right;
		unlock(&lparent->lock);
		unlock(&sparent->lock);
		unlock(&successor->lock);
		swap_node(new, previous, &node);
		if(pivotal) {
			pivotal->node.ptr = new->node.ptr;
			unlock(&pivotal->lock);
			new->node = node;
			unlock(&new->lock);
			entry = pivotal;
		}
		else {
			new->node.address = address;
			entry = new;
		}
		device_write(cache->fs->device, entry->node.address, entry->node.ptr, cache->fs->size.block);
	}
	/* vcp is virtual copy, buffer is a physical address, a RAM location;
	 * while ptr is a virtual address. The opposite of vcp is pcpy. */
	vcp(entry->node.ptr, buffer, cache->fs->size.block);
	adjust_entry(cache, entry);
	return cache;
}
