#ifndef CHUNK_H
#define CHUNK_H

/*
    Copyright (C) 2006 Sundaresh Venugopal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    sundaresh@mail.com
*/

#include<sys/time.h>

#include"list.h"
#include"partial.h"

enum chunk_type { CLUSTERED, SECTIONED };

/* orphaned nodes */

struct chunk_entry {
	union chunk_base {
		unsigned long clustered;
		struct sectioned_base {
			unsigned section;
			unsigned offset;
		} stretch;
	} base;
	struct chunk_length {
		unsigned total;
		unsigned used;
	} length;
};

struct chunk {
	struct chunk_entry entry;
	list_ptr_t recent;
};

union chunk_node {
	struct chunk chunk;
	union chunk_node *next;
};

struct chunk_file_table {
	struct chunk *base;
	table_t memory, data;
	list_t least;
};

struct chunk_toc_table {
	struct partial_table disk, core;
	struct partial_table pieces;
};

/* chunk on-disk inode */
struct chunk_inode {
	struct partial *name;
	int type;
	short link;
	short owner, group;
	short permission;
	long size;
	struct file_time {
		struct timeval created, accessed, read, written;
	} time;
	union directory_ptr {
		struct chunk_inode *dir, *link;
	} ptr;
	struct partial *toc, *pieces;
};

union chunk_inode_node {
	struct chunk_inode chunk_inode;
	union chunk_inode_node *next;
};

struct chunk_inode_table {
	table_t table;
	union chunk_inode_node *base, *avail;
	lock_t free;
};

struct chunk_fs_table {
	void *namebase;
	struct partial_table name;
	struct chunk_file_table file;
	struct chunk_toc_table toc;
	struct chunk_inode_table inode;
};

struct chunk_toc_ptr {
	bool_t pieces;
	short offset;
	struct partial *partial;
};

struct chunk_toc {
	struct partial *whole, *pieces;
	struct chunk_toc_ptr ptr;
};

#endif
