#ifndef FILE_H
#define FILE_H

/*
    Copyright (C) 2006  Sundaresh Venugopal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    sundaresh@mail.com
*/

#include<sys/time.h>

#include"fs.h"

#include"chunk.h"

#define NAME_LEN (32)

struct fs;

struct file_info {
	int type;
	short link;
	char name[NAME_LEN];
	short owner, group;
	short permission;
	long length;
	struct file_time {
		struct timeval created, accessed, read, written;
	} time;
};

union file_toc {
	struct chunk_toc chunk;
};

struct file_pos {
	long pos;
	long address;
	int count;
};

/* on-disk inode */
union file_inode {
	struct chunk_inode *chunk;
};

/* in-core inode */
struct inode {
	struct file_info info;
	union file_toc toc;
	struct file_pos pos;
	union file_inode disk;
};

struct file {
	int count;
	struct fs *fs;
	struct file *dir;
	struct inode inode;
	list_t clist;
};

#define ismount(file) ((file)->fs->root == (file))

struct file_ops {
	struct inode *(*open)(struct fs *fs, char *relative, int mode);
	long (*pos)(struct inode *inode, long *pos);
	long (*length)(struct inode *inode, long length);
	struct inode *(*sync)(struct inode *inode);
};

#endif
