
#include<sys/types.h>

#include"../../include/vm.h"

struct frame {
	struct frame_flags {
		unsigned char dirty : 1;
		unsigned char locked : 1;
		unsigned char enabled : 1;
		unsigned char disabled : 1;
	} flag;
	unsigned page;
	struct frame *previous, *next;
	struct frame *left, *right;
	struct frame_usage {
		unsigned age;
		short level;
		struct frame *previous, *next;
	} usage;
	unsigned short read[2], write[2];
};
#define FRAME_SIZE (sizeof(struct frame))

struct frame_table {
	unsigned age;
	struct frame *frame, *head[VM_USAGE_LEVELS + 1], *least;
};

struct fix_table {
	void *base;
	table_t association;
};

struct translation_table {
	struct fix_table prefix, suffix;
	struct frame_table frame;
};

struct request {
	unsigned page;
	unsigned short count : 15;
	unsigned short complete : 1;
	void *frame;
};
#define REQUEST_SIZE (sizeof(struct vm_request))

union request_node {
	struct request request;
	union request_node *next;
};

struct request_table {
	union request_node base[VM_REQUEST_MAX], *avail, *last;
	table_t table;
};

struct vm_mutex {
	pthread_mutex_t realloc, sync, load, lru, lav, head, avail;
};

enum vm_flags { VM_INHERIT = 1, VM_REALLOC };

struct vm {
	pid_t process;
	unsigned flag;
	int _virtual;
	unsigned pages, last, limit;
	short access;
	struct request_table request;
	struct vm_mutex mutex;
};
#define VM_SIZE (sizeof(struct vm))

union vm_node {
	struct vm space;
	union vm_node *next;
};

struct vm_pool {
     union vm_node base[VM_MAX], *avail;
};

typedef const int vm_t;

struct vms_mutex {
	pthread_mutex_t realloc, sync, request, lru, lav, head, avail;
};

struct vms_stat;

struct vms_page {
	unsigned char page[VM_PAGE_SIZE];
	struct vms_page *next;
};

struct vms_cache_entry {
	unsigned page;
	struct vms_page *ptr;
};

struct vms_cache {
	vm_t memory;
	struct vms_page *base, *first, *last, *head[VM_MAX];
	table_t table[VM_MAX];
};

struct vms {
	struct physical {
		void *base;
		int frames, length[3];
		struct frame_table ;
	
	struct frame_table frame;
	struct vm_pool pool;
	struct vms_cache cache;
	struct vms_mutex mutex;
	struct vms_stat stat;
} vms;
