
#include<sys/types.h>

#include"../../include/vm.h"

struct frame {
	struct frame_flags {
		unsigned char dirty : 1;
		unsigned char enabled : 1;
		unsigned char disabled : 1;
		unsigned char locked : 1;
	} flag;
	unsigned page;
	struct frame *previous, *next;
	struct frame *left, *right;
	struct frame_usage {
		unsigned short level;
		struct frame *previous, *next;
	} usage;
	unsigned short read[2], write[2];
};
#define FRAME_SIZE (sizeof(struct frame))

struct fault {
	unsigned page;
	unsigned short count;
	void *frame;
	struct fault *left, *right;
};
#define FAULT_SIZE (sizeof(struct fault))

union fault_node {
	struct fault fault;
	union fault_node *next;
};

struct level {
	struct frame *head;
	unsigned short previous, next;
};

union level_node {
	struct level level;
	union level_node *next;
};

struct cache {
	unsigned page;
	void *ptr;
	struct cache *previous, *next;
	struct cache *left, *right;
	struct cache_usage {
		struct cache *previous, *next;
	} usage;
};

struct vm_root {
	struct frame *frame;
	struct fault *fault;
};

struct vm_mutex {
	pthread_mutex_t realloc, sync, fault, least, head, avail;
};

struct vm {
	pid_t creator;
	unsigned frames;
	unsigned pages, last;
	short access;
	struct frame *head;
	struct vm_root root;
	struct vm_mutex mutex;
};

union vm_node {
	struct vm space;
	union vm_node *next;
};

struct vms_table {
	struct frame *frame;
	struct usage_table {
		union level_node level[LEVEL_MAX], *avail;
		struct frame *least;
	} usage;
	struct fault_table {
		union fault_node base[FAULT_MAX], *avail;
		struct fault *least;
	} fault;
	struct cache_table {
		int memory;
		struct cache base[CACHE_MAX];
		struct cache *least, *most;
	} cache;
};

struct vms_pool {
     union vm_node base[VM_MAX], *avail;
};

struct vms_stat;

struct vms {
	void *physical;
	unsigned frames;
	struct vms_table table;
	struct vms_pool pool;
	struct vms_stat stat;
};

extern struct vms vms;
