#ifndef VMXDEF_H
#define VMXDEF_H

#include<sys/types.h>

#include"../../include/vm.h"

struct vm_frame {
	unsigned page : 31;
	unsigned dirty : 1;
	unsigned enable : 1;
	unsigned disable : 1;
	unsigned counter : 24;
	unsigned usage : 6;
	unsigned short read[2], write[2];
	struct vm_frame *next;
	struct {
	   struct vm_frame *left, *right;
	} least;
};
#define VM_FRAME_SIZE (sizeof(struct vm_frame))

struct vm_frame_table {
	struct vm_frame *base, **head;
	struct vm_frame_usage {
		struct vm_frame *root, *least;
	} usage;
};

struct vm_fix {
	unsigned short count;
	struct vm_fix *left, *right;
};
#define VM_FIX_SIZE (sizeof(struct vm_fix))

struct vm_fix_table {
	void *base;
	struct vm_fix_association {
		struct vm_fix *base;
		struct vm_fix *root, *least;
	} association;
};

struct vm_request {
	unsigned page : 31;
	unsigned complete : 1;
	unsigned short count;
	void *frame;
	struct vm_request *previous, *next;
};
#define VM_REQUEST_SIZE (sizeof(struct vm_request))

struct vm_request_table {
	struct vm_request *base, *avail, *last;
	unsigned short *search;
};

struct vm_table {
	struct vm_fix_table prefix, suffix;
	struct vm_frame_table frame;
	struct vm_request_table request;
};

struct vm_mutex {
	pthread_mutex_t realloc, sync, load, lru, lav, head, avail;
};

struct vm_stat;

enum vm_flags { VM_REALLOC = 1, VM_SYNC };

struct vm {
	char flags;
	int _virtual;
	void *physical;
	int pages, last;
   	int frames;
	short access;
	struct vm_table table;
	struct vm_mutex mutex;
	struct vm_stat stat;
};
#define VM_SIZE (sizeof(struct vm))

#endif
