#ifndef COUNTER_H
#define COUNTER_H

/*
    Copyright (C) 2006  Sundaresh Venugopal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    sundaresh@mmail.com
*/

/* Author : sundaresh venugopal */

#include<time.h>

#ifndef EACTIVE
#define EACTIVE
#endif

enum op_type { PARALLEL, SEQUENTIAL };

struct op_counter {
	unsigned up, down;
};

enum counter_type { BALANCED, BIASED_PARALLEL = 1, BIASED_SEQUENTIAL = 3 };

struct balanced_counter {
	struct op_counter parallel, sequential;
};

struct biased_counter {
	unsigned parallel, sequential;
};

#ifndef COUNTER_T
#define COUNTER_T
typedef struct counter {
	unsigned char rwlock : 1;
	unsigned char type : 7;
        union {
		struct balanced_counter balanced;
		struct biased_counter biased;
	} counter;
} counter_t;
#endif

void counter_pool_init();

void counter_init(counter_t *counter, int type);
int counter_tryup(counter_t *counter, int op);
int counter_timedup(counter_t *counter, int op, struct timespec *abstime);
#define counter_up(counter, op) (counter_timedup(counter, op, NULL))
int counter_down(counter_t *counter, int op);
int counter_active(counter_t *counter);

#endif
