
#include<pthread.h>
#include<signal.h>
#include<errno.h>

#include"list.h"

#include"lock.h"
#include"deadlock.h"

void lock_init(lock_t *lock) {
	pthread_cancel_t cancel;
	pthread_atomic_t atomic;

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	pthread_setatomicstate(PTHREAD_ATOMIC, &atomic);
	lock->thread = NULL;
	lock->lock = lock->mutex = 0;
	listptr_init(&lock->ptr.lock, LOCK_LIST_TYPE);
	listptr_init(&lock->ptr.deadlock, LOCK_LIST_TYPE);
	pthread_setcancelstate(cancel, NULL);
	pthread_setatomicstate(atomic, NULL);
	return;
}

int trylock(lock_t *lock) {
	pthread_t self;
	int success;

	if(lock->thread == (self = pthread_self())) success = 1;
	else {
		if(! (lock->thread || lock->lock)) {
			pthread_cancel_t cancel;

			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
			lock->lock++;
			if(lock->lock == 1) {
				if(lock->mutex || list_insert(self->list.lock, lock, &lock->ptr.lock)) {
					lock->thread = self;
					if(! self->process.deadlock_check ||
						list_delete(self->list.deadlock, &lock->ptr.deadlock)) {
						success = 1;
					}
					else success = 2;
				}
			}
			else {
				lock->lock--;
				success = -1;
			}
			pthread_setcancelstate(cancel, NULL);
		}
		else success = 0;
		errno = ((lock->thread->process->deadlock_check) && deadlocked(lock)) ? EDEADLK : EBUSY;
	}
	return success;
}

int lock(lock_t *lock) {
	unsigned retries = 0, counter = 0, start = 0;

	while(trylock(lock) <= 0) {
		if(errno == EDEADLK) {
			if(counter != ++retries) start = retries = counter;
			else {
				if(retries - start >= DEADLOCK_RETRIES) raise(SIG_DEADLK);
			}
		}
		pthread_switch();
		++counter;
	}
	return 1;
}

int unlock(lock_t *lock) {
	pthread_t self;
	int success;

	if(lock->thread == (self = pthread_self()) && lock->lock == 1) {
		pthread_cancel_t cancel;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		lock->lock--;
		lock->thread = NULL;
		if(lock->mutex || list_delete(self->list.lock, &lock->ptr.lock)) success = 1;
		else success = -1;
		pthread_setcancelstate(cancel, NULL);
	}
	else success = 0;
	return success;
}
