
#include<signal.h>

#include"lock.h"
#include"mutex.h"

int pthread_mutex_init(pthread_mutex_t *mutex, pthread_mutexattr_t *attr) {
	pthread_cancel_t cancel;
	int len;

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	if(mutex->dead) {
		lock_init(&mutex->lock);
		mutex->lock.mutex = 1;
		if(lock(&mutex->lock) > 0) {
			if(mutex->dead) {
				mutex->dead = FALSE;
				len = *attr && *attr > 0 ? *attr : PTHREAD_MUTEX_QUEUE_LEN;
				mutex->queue = list_create(QUEUE_TYPE, len);
			}
			unlock(&mutex->lock);
			pthread_setcancelstate(cancel, NULL);
			return (mutex->queue) ? 0 : -1;
		}
	}
	pthread_setcancelstate(cancel, NULL);
	return -2;
}

int pthread_mutex_destroy(pthread_mutex_t *mutex) {
	pthread_cancel_t cancel;

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	if(! pthread_mutex_lock(mutex)) {
		mutex->dead = TRUE;
		list = mutex->list;
		mutex->list = NULL;
		thread = list_get(mutex->list, GET_FIRST, NULL);
		while(thread) {
			pthread_kill(thread, SIG_RESUME);
			thread = list_get(mutex->list, GET_NEXT, NULL);
		}
		list_destroy(mutex->list);
		mutex->list = NULL;
		unlock(&mutex->lock);
		pthread_setcancelstate(cancel, NULL);
		return 0;
	}
	pthread_setcancelstate(cancel, NULL);
	return 1;
}

/* thread is relative to kernel data space, while mutex is relative to process data space */
int pthread_mutex_trylock(pthread_mutex_t *mutex) {
	pthread_atomic_t atomic;
	pthread_cancel_t cancel;
	pthread_t self;
	int success;

	pthread_setatomicstate(PTHREAD_ATOMIC, &atomic);
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	if(! mutex->dead) {
		self = pthread_self();
		if(! mutex->queue) {
			if(! mutex->queue = list_create(QUEUE_TYPE, PTHREAD_MUTEX_QUEUE_LEN)) {
				success = -1;
			}
		}
		pthread_setatomicstate(atomic, NULL);
		if(! list_count(mutex->queue)) {
			if(trylock(&mutex->lock) > 0) {
				mutex->ptr.linked.double = NULL;
				if(list_insert(self->list.mutex, mutex, &mutex->lock.ptr.lock)) {
					success = 0;
				}
				unlock(&mutex->lock);
				success = -1;
			}
		}
		success = 1;
	}
	else success = -2;
	pthread_setcancelstate(cancel, NULL);
	return success;
}

/* pthread_suspend suspends execution of the thread until mutex becomes available to it */
static int pthread_suspend(pthread_mutex_t *mutex) {
	pthread_cancel_t cancel;
	pthread_t self;
	list_t queue;
	int success;

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	if(! mutex->dead) {
		self = pthread_self();
		if(queue = mutex->queue) {
			if(list_append(queue, self, NULL)) {
				raise(SIG_SUSPEND);
				if(trylock(&mutex->lock) > 0) success = 0;
			}
			else success = -1;
		}
	}
	else success = -2;
	pthread_setcancelstate(cancel, NULL);
	return success;
}

int pthread_mutex_lock(pthread_mutex_t *mutex) {
	if(pthread_mutex_trylock(mutex)) return pthread_suspend(mutex);
}

int pthread_mutex_unlock(pthread_mutex_t *mutex) {
	pthread_cancel_t cancel;
	pthread_t self;

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
	if(! mutex->dead) {
		if(mutex->lock.thread == (self = pthread_self())) {
			pthread_t next;
			list_t queue;

			if(list_delete(self->list.mutex, &mutex->ptr)) {
				if(queue = mutex->queue) {
					if(next = list_get(queue, GET_FIRST, NULL)) {
						if(! list_execute(queue, NULL)) success = -1;
					}
					unlock(&mutex->lock);
					pthread_kill(next, SIG_RESUME);
					success = 0;
				}
			}
			else success = -1;
		}
		else success = 1;
	}
	else success = -2;
	pthread_setcancelstate(cancel, NULL);
	return success;
}
