
#include<pthread.h>

#include"lock.h"
#include"counter.h"
#include"deadlock.h"

int deadlocked(lock_t *lock) {
	/* thread is relative to kernel data space, rest are relative to user data space */
	pthread_t thread, sequence[PTHREADS_MAX];
	table_t deadlock;
	struct deadlock *request;
	int pos, last;

	thread = pthread_self();
	counter_up(&thread->counter.deadlock, PARALLEL);
	request = thread->head.deadlock;
	while(request && request->thread != lock->thread) request = request->next;
	counter_down(&thread->counter.deadlock, PARALLEL);
	if(! request) {
		struct deadlock *avail, *most;

		counter_up(&thread->counter.deadlock, SEQUENTIAL);
		while(lock(&thread->process->pool.deadlock->used.lock) <= 0) pthread_skip();
		avail = thread->process->pool.deadlock->used.least;
		thread->process->pool.deadlock->used.least = avail->least.previous;
		avail->least.previous = NULL;
		avail->least.next = most = thread->process->pool.deadlock->used.most;
		most->least.previous = thread->process->pool.deadlock->used.most = avail;
		unlock(&thread->process->pool.deadlock->used.lock);
		avail->thread = lock->thread;
		avail->next = thread->head.deadlock;
		thread->head.deadlock = avail;
		counter_down(&thread->counter.deadlock, SEQUENTIAL);
	}
	pos = last = 0;
	deadlock = table_create(KERNEL_TABLE + NO_DUPLICATES + LOCAL_TABLE + DEADLOCK_TABLE, PTHREAD_MAX, NULL, NULL);
	entry->thread = pthread_self();
	sequence[last] = lock->thread;
	while(pos <= last) {
		thread = sequence[pos];
		counter_up(&thread->counter.deadlock, PARALLEL);
		request = thread->deadlock.head;
		while(request) {
			last++;
			sequence[last] = request->thread;
			if(! table_insert(deadlock, entry->thread)) {
				counter_down(&thread->counter.deadlock, PARALLEL);
				return 1;
			}
			request = request->next;
		}
		counter_down(&thread->counter.deadlock, PARALLEL);
		pos++;
	}
	table_destroy(deadlock);
	return 0;
}
