
#include<asm/atomic.h>

#include<errno.h>

#include"lock.h"
#include"counter.h"

#ifndef NDEADLOCK_ADVISORY
#include"deadlock.h"
#endif

int lock(lock_t *lock) {
	if(! (lock->thread || atomic_read(&lock->lock))) {
		int cancel;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		atomic_inc(&lock->lock);
		if(atomic_read(&lock->lock) == 1) {
			lock->thread = pthread_self();
#ifndef NDEADLOCK_ADVISORY
			if(lock->thread->process.deadlock_check) {
				struct deadlock *previous, *request;
				pthread_t thread;

				thread = lock->thread;
				counter_up(&thread->counter.deadlock, PARALLEL);
				previous = NULL;
				request = thread->head.deadlock;
				while(request && request->thread != thread) {
					previous = request;
					request = request->next;
				}
				counter_down(&thread->counter.deadlock, PARALLEL);
				if(request) {
					counter_up(&thread->counter.deadlock, SEQUENTIAL);
					if(previous) previous->next = request->next;
					else thread->deadlock.head = request->next;
					while(lock(&thread->process->pool.deadlock->used.lock) <= 0) pthread_skip();
					if(request->used.next) {
						request->used.next->used.previous = request->used.previous;
						request->used.next = least = thread->process->pool.deadlock->used.least;
						thread->process->pool.deadlock->used.least = least->previous = request;
					}
					if(request->used.previous) {
						request->used.previous->used.next = request->used.next;
						request->used.previous = NULL;
					}
					else thread->process->pool.deadlock->used.most = request->used.next;
					unlock(&thread->process->pool.deadlock->used.lock);
					counter_down(&thread->counter.deadlock, SEQUENTIAL);
				}
			}
#endif
			pthread_setcancelstate(cancel, NULL);
			return 1;
		}
		atomic_dec(&lock->lock);
		pthread_setcancelstate(cancel, NULL);
		return 0;
	}
	if(lock->thread == pthread_self()) return 1;
#ifndef NDEADLOCK_ADVISORY
	errno = lock->thread->process->deadlock_check && deadlocked(lock) ? EDEADLK : EBUSY;
#else
	errno = EBUSY;
#endif
	return -1;
}

int unlock(lock_t *lock) {
	if(lock->thread == pthread_self() && atomic_read(&lock->lock) == 1) {
		int cancel;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		atomic_dec(&lock->lock);
		lock->thread = NULL;
		pthread_setcancelstate(cancel, NULL);
		return 1;
	}
	return 0;
}
