
#include<asm/atomic.h>

#include<errno.h>

#include"list.h"

#include"lock.h"
#include"counter.h"

#include"deadlock.h"

int trylock(lock_t *lock) {
	pthread_t self;

	if(lock->thread == (self = pthread_self())) return 1;
	if(! (lock->thread || atomic_read(&lock->lock))) {
		int cancel;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		atomic_inc(&lock->lock);
		if(atomic_read(&lock->lock) == 1) {
			pthread_t thread;
			list_ptr_t ptr;

			thread = lock->thread = self;
			list_insert(self->list.lock, thread, NULL);
			if(self->process.deadlock_check) {
				thread = list_get(self->list.deadlock, GET_FIRST, &ptr);
				while(thread && thread != self) {
					thread = list_get(self->list.deadlock, GET_NEXT, &ptr);
				}
				list_delete(self->list.deadlock, &ptr);
			}
			pthread_setcancelstate(cancel, NULL);
			return 1;
		}
		atomic_dec(&lock->lock);
		pthread_setcancelstate(cancel, NULL);
	}
	errno = EBUSY;
	return 0;
}

int lock(lock_t *lock) {
	retries = LOCK_RETRIES;
	while(retries && trylock(lock) <= 0) {
		if(lock->thread->process->deadlock_check && deadlocked(lock)) {
			errno = EDEADLK;
			--retries;
		}
		pthread_skip();
	}
	return retries ? 0 : 1;
}

int unlock(lock_t *lock) {
	if(lock->thread == pthread_self() && atomic_read(&lock->lock) == 1) {
		int cancel, thread, self;
		list_ptr_t ptr;

		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cancel);
		atomic_dec(&lock->lock);
		lock->thread = NULL;
		self = pthread_self();
		thread = list_get(self->list.lock, GET_FIRST, &ptr);
		while(thread != self) {
			thread = list_get(self->list.lock, GET_NEXT, &ptr);
		}
		list_delete(self->list.lock, &ptr);
		pthread_setcancelstate(cancel, NULL);
		return 1;
	}
	return 0;
}
