
#include<errno.h>

#include"counter.h"
#include"rwlock.h"
#include"mutex.h"

int pthread_rwlock_init(pthread_rwlock_t *rwlock, pthread_rwlockattr_t *attr) {
	rwlock->type = *attr;
	counter_init(&rwlock->counter, 2 * *attr);
	return 0;
}

int pthread_rwlock_destroy(pthread_rwlock_t *rwlock) {
	if(counter_active(&rwlock->counter, 2 * rwlock->type)) return errno = EBUSY;
       	return 0;
}

static struct rwlock_entry *previous;

static struct rwlock_entry *find_rwlock(pthread_rwlock_t *rwlock) {
	struct rwlock_entry *request;
	pthread_t thread;

	thread = pthread_self();
	previous = NULL;
	request = thread->head.rwlock;
	while(request && request->rwlock != rwlock) {
		previous = request;
		request = request->next;
	}
	return request;
}

static int add_rwlock(pthread_rwlock_t *rwlock, bool_t write) {
	struct rwlock_entry *request, *new;
	union rwlock_node *avail;
	pthread_t thread;

	thread = pthread_self();
	while(lock(&thread->process->pool.rwlock->free) <= 0) pthread_skip();
	avail = thread->process->pool.rwlock->avail;
	thread->process->pool.rwlock->avail = avail.next;
	unlock(&thread->process->pool.rwlock->free);
	new = &avail.rwlock;
	new->write = write;
	new->rwlock = rwlock;
	new->next = thread->head.rwlock;
	thread->head.rwlock = new;
	counter_up(&rwlock->counter, (write ? SEQUENTIAL : PARALLEL) | (2 * rwlock->type));
	return 1;
}

int pthread_rwlock_trylock(pthread_rwlock_t *rwlock, bool_t write) {
	if (! counter_active(&rwlock->counter, 2 * rwlock->type)) return add_rwlock(rwlock, write);
	errno = EBUSY;
	return 0;
}

int pthread_rwlock_lock(pthread_rwlock_t *rwlock, bool_t write) {
	struct rwlock_entry *entry;

	if(! (entry = find_rwlock(rwlock))) return add_rwlock(rwlock, write);
	if(! write && entry->write) return 1;
	errno = EBUSY;
	return 0;
}

int pthread_rwlock_unlock(pthread_rwlock_t *rwlock) {
	struct rwlock_entry *entry;
	union rwlock_node *avail;
	pthread_t thread;

	if(entry = find_rwlock(rwlock)) {
		thread = pthread_self();
		counter_down(&entry->rwlock->counter, (entry->rwlock->type * 2) | entry->write ? SEQUENTIAL : PARALLEL);
		if(previous) previous->next = entry->next;
		else thread->head.rwlock = entry->next;
		pthread_mutex_lock(&thread->process->pool.rwlock.free);
		avail = (union rwlock_node *)entry;
		avail.next = thread->process->pool.rwlock.avail;
		thread->process->pool.rwlock->avail = avail;
		pthread_mutex_unlock(&thread->process->pool.rwlock->free);
	}
	return 1;
}
